#pragma once
#include "mlp.h"

#include <chrono>
#include <algorithm>
#include <numeric>

enum class DECrossoverStrategy {
	Rand_1,			//DE/rand/1
	Best_1,			//DE/best/1
	Cur_To_Rand_1,	//DE/cur-to-rand/1
	Cur_To_Best_1,	//DE/cur-to-best/1
	Cur_To_PBest_1,	//DE/cur-to-pbest/1 (JADE)

	count,
};

constexpr size_t RandomMode = 999;

struct Genome {
	std::unique_ptr<NETWORK> n, next_n;
	double fitness = std::numeric_limits<double>::quiet_NaN(), next_fitness = std::numeric_limits<double>::quiet_NaN();
	double diff_F = 0.1;
	double diff_CR = 0.5;
	size_t diff_P = 5;
	DECrossoverStrategy diff_strategy = DECrossoverStrategy::Rand_1;
	size_t batch_size = 1;
	size_t batch_start = 0;
};

struct Genome_Backprop : public Genome {
	double backprop_rate = 0.01;
	double backprop_momentum = 0.01;
	size_t backprop_epochs = 10;
	double bp_fitness = 1.0;
};

static size_t getWeightCount(std::unique_ptr<NETWORK>& n) {
	size_t wcount = n->inputs + 1;
	for (size_t j = 0; j < n->layers.size(); j++) {
		if (j == 0)
			wcount *= n->layers[0].neurons.size();
		else
			wcount += (n->layers[j - 1].neurons.size() + 1) * n->layers[j].neurons.size();
	}

	return wcount;
}

template<bool Condition, typename T, typename U>
struct static_type_switch {};

template<typename T, typename U>
struct static_type_switch<true, T, U> { typedef T type; };

template<typename T, typename U>
struct static_type_switch<false, T, U> { typedef U type; };

template<typename FType>
bool selectRand(double randNum, double chance1, FType&& func1) {
	if (randNum < chance1) {
		func1();
		return true;
	}

	return false;
}

template<typename FType, typename ...Tail>
bool selectRand(double randNum, double chance1, FType&& func1, Tail&&... tail) {
	if (randNum < chance1) {
		func1();
		return true;
	}

	return selectRand(randNum - chance1, std::forward<Tail>(tail)...);
}

template<bool WithBackprop, bool WithEvolution>
double genetic_algorithm(std::unique_ptr<NETWORK>& n, const std::vector<std::vector<double>>& sample, size_t samples, const std::vector<std::vector<double>>& desired, size_t population_size_request = 40, size_t generations_count_request = 1000) {

	using Genome_Type = typename static_type_switch<WithBackprop, Genome_Backprop, Genome>::type;

	const size_t population_size = population_size_request;
	const size_t generations_count_min = generations_count_request;

	const double p_crossover_weights =	WithBackprop ? 0.05  : 0.15;
	const double p_crossover_pathway =	WithBackprop ? 0.05  : 0.15;
	const double p_crossover_diff =		WithBackprop ? 0.3   : 0.35;
	const double p_mutation =			WithBackprop ? 0.1   : 0.1;
	const double p_backprop =			WithBackprop ? 0.4   : 0;
	const double p_meta_crossover =		WithBackprop ? 0.1   : 0;
	const double p_blitz =				WithBackprop ? 0.002 : 0.002;
	const double p_freely =				WithBackprop ? 0.002 : 0.002;
	const double p_randomgen =			WithBackprop ? 0.002 : 0.01;
	const double p_elitereinit =		WithBackprop ? 0.01  : 0.01;
	const double p_rate_change =		WithBackprop ? 0.2   : 0.15;

	const double diff_jade_c = 0.5;

	std::vector<Genome_Type> population;
	Genome_Type elite;
	double lastSavedFitness = std::numeric_limits<double>::max();

	std::random_device rdev;
	std::default_random_engine reng(rdev());

	std::uniform_int_distribution<size_t> layer_gen(0, n->layers.size() - 1);
	std::uniform_int_distribution<size_t> pop_gen(0, population_size - 1);
	std::uniform_real_distribution<double> norm_gen(0.0, 1.0);

	std::normal_distribution<double> norm_BPRate_gen(0.2, 0.5);
	std::normal_distribution<double> norm_Momentum_gen(0.2, 0.5);

	std::normal_distribution<double> norm_CR_gen(0.5, 0.1);
	std::cauchy_distribution<double> cauchy_F_gen(0.5, 0.1);

	auto genNumber = [&norm_gen, &reng](size_t from, size_t to) {
		return from + static_cast<size_t>(static_cast<double>(to - from) * norm_gen(reng));
	};

	auto genBackpropRate = [&norm_gen, &reng, &norm_BPRate_gen]() {

		double rate;

		do {
			rate = norm_BPRate_gen(reng);
		} while (rate < std::numeric_limits<double>::epsilon());

		return std::max(0.0, std::min(1.0, rate)); //std::pow(norm_gen(reng) * 0.5, 2.0);
	};

	auto genMomentumRate = [&norm_gen, &reng, &norm_Momentum_gen]() {

		double rate;

		do {
			rate = norm_Momentum_gen(reng);
		} while (rate < std::numeric_limits<double>::epsilon());

		return std::max(0.0, std::min(1.0, rate)); //std::pow(norm_gen(reng) * 0.5, 2.0);
	};

	auto genBackpropEpochs = std::bind(genNumber, 2, 20);

	auto genBatchSize = [population_size, samples, &genNumber]() {
		return WithBackprop ? genNumber((80 * samples) / 100, samples) : samples; //velikost davky je 80 - 100 %
	};

	auto genBatchStart = [population_size, samples, &genNumber](size_t batchSize) {
		return WithBackprop ? genNumber(0ULL, samples - batchSize) : 0;
	};

	auto genFParameter = [&cauchy_F_gen, &reng]() {

		double F;

		do {
			F = cauchy_F_gen(reng);
		} while (F < std::numeric_limits<double>::epsilon());

		return std::min(1.0, std::max(0.0, F));
	};

	auto genCRParameter = [&norm_CR_gen, &reng]() {
		return std::min(1.0, std::max(0.0, norm_CR_gen(reng)));
	};

	auto genPBestParameter = [&genNumber, population_size]() {
		return population_size > 5 ? genNumber(5, population_size / 2) : 5;
	};

	auto genDEStrategy = [&genNumber]() {
		return static_cast<DECrossoverStrategy>(genNumber(0, static_cast<size_t>(DECrossoverStrategy::count)));
	};


	//lambda funkce - inicializace clena populace strategii blitz
	auto blitz = [&sample, &desired, &genNumber](Genome_Type& trg, size_t mode = RandomMode) {
		blitzInit(trg.next_n, sample, desired, (mode == RandomMode) ? genNumber(1, 3) : mode);

		trg.fitness = std::numeric_limits<double>::max();
	};

	//lambda funkce - inicializace clena populace strategii freely growing
	auto freelyGrow = [&sample, &desired, &genNumber](Genome_Type& trg, size_t mode = RandomMode) {
		freelyGrowingInit(trg.next_n, sample, desired, (mode == RandomMode) ? genNumber(1, 4) : mode);

		trg.fitness = std::numeric_limits<double>::max();
	};

	// lambda funkce - nahodna reinicializace
	auto randomReinit = [](Genome_Type& trg) {

		std::unique_ptr<NETWORK>& n = trg.n;

		for (size_t layer = 0; layer < n->layers.size(); layer++) {
			for (size_t neuron = 0; neuron < n->layers[layer].neurons.size(); neuron++) {
				for (size_t weight = 0; weight < n->layers[layer].neurons[neuron].weights.size(); weight++) {
					n->layers[layer].neurons[neuron].weights[weight] = drand();
					trg.next_n->layers[layer].neurons[neuron].weights[weight] = n->layers[layer].neurons[neuron].weights[weight];
				}
			}
		}

		trg.fitness = std::numeric_limits<double>::max();
	};

	//vygeneruje se pocatecni populace
	if constexpr (WithBackprop) {
		for (size_t i = 0; i < population_size - 1; i++) {
			population.push_back({ cloneNetwork(n, true), cloneNetwork(n, true), std::numeric_limits<double>::quiet_NaN(), std::numeric_limits<double>::quiet_NaN(), genFParameter(), genCRParameter(), genPBestParameter(), genDEStrategy(), genBatchSize(), 0, genBackpropRate(), genBackpropRate(), genBackpropEpochs()});
		}
		population.push_back({ cloneNetwork(n, false), cloneNetwork(n, false), std::numeric_limits<double>::quiet_NaN(), std::numeric_limits<double>::quiet_NaN(), genFParameter(), genCRParameter(), genPBestParameter(), genDEStrategy(), genBatchSize(), 0, genBackpropRate(), genBackpropRate(), genBackpropEpochs() });
	}
	else {
		for (size_t i = 0; i < population_size - 1; i++) {
			population.push_back({ cloneNetwork(n, true), cloneNetwork(n, true), std::numeric_limits<double>::quiet_NaN(), std::numeric_limits<double>::quiet_NaN(), genFParameter(), genCRParameter(), genPBestParameter(), genDEStrategy(), genBatchSize(), 0 });
		}
		population.push_back({ cloneNetwork(n, false), cloneNetwork(n, false), std::numeric_limits<double>::quiet_NaN(), std::numeric_limits<double>::quiet_NaN(), genFParameter(), genCRParameter(), genPBestParameter(), genDEStrategy(), genBatchSize(), 0 });
	}
	elite.n = cloneNetwork(n, false);

	//lambda funkce - vygenerovani casti populace podle nejakeho klice
	auto initPopPart = [&population, population_size](size_t partIdx, size_t partsTotal, std::function<void(Genome_Type&)> callback) {

		size_t partSize = population_size / partsTotal;

		for (size_t i = partIdx * partSize; i < (partIdx + 1) * partSize; i++) {
			callback(population[i]);
		}
	};

	//lambda funkce - prepocitani fitness funkce vsech v populaci
	auto calcPopulationFitness = [&sample, &samples, &desired, &population, population_size]() {
		std::for_each(std::execution::par_unseq, population.begin(), population.end(), [&sample, &samples, &desired](Genome_Type& g) {
			g.next_fitness = avgQuadError(g.next_n, sample, 0, samples, desired);
		});
	};

	//vyber - jen lepsi preziji
	auto selection = [&population]() {
		std::for_each(std::execution::par_unseq, population.begin(), population.end(), [](Genome_Type& g) {
			if (g.next_fitness < g.fitness) {
				g.fitness = g.next_fitness;
				cloneNetworkWeights(g.next_n, g.n);
			}
			else {
				cloneNetworkWeights(g.n, g.next_n);
			}
		});
	};

	//serazeni populace
	auto sortPopulation = [&population, population_size]() {
		std::partial_sort(std::execution::par_unseq, population.begin(), population.begin() + (population_size / 2), population.end(), [](const Genome_Type& a, const Genome_Type& b) {
			return a.fitness < b.fitness;
		});
	};

	//lambda funkce - ulozeni elitniho clena populace do promenne
	auto saveElite = [&elite, &population]() {
		cloneNetworkWeights(population[0].n, elite.n);
		elite.diff_F = population[0].diff_F;
		elite.diff_P = population[0].diff_P;
		elite.diff_CR = population[0].diff_CR;
		elite.fitness = population[0].fitness;
		elite.batch_start = population[0].batch_start;
		elite.batch_size = population[0].batch_size;
		if constexpr (WithBackprop) {
			elite.backprop_epochs = population[0].backprop_epochs;
			elite.backprop_rate = population[0].backprop_rate;
			elite.bp_fitness = population[0].bp_fitness;
			elite.backprop_momentum = population[0].backprop_momentum;
		}
	};

	// lambda funkce - obnoveni elitniho clena populace zpatky do populace na zadane misto
	auto restoreElite = [&elite, &population, samples](size_t target) {
		cloneNetworkWeights(elite.n, population[target].n);
		population[target].diff_F = elite.diff_F;
		population[target].diff_P = elite.diff_P;
		population[target].diff_CR = elite.diff_CR;
		population[target].fitness = elite.fitness;
		population[target].batch_start = elite.batch_start;
		population[target].batch_size = elite.batch_size;
		if constexpr (WithBackprop) {
			population[target].backprop_epochs = elite.backprop_epochs;
			population[target].backprop_rate = elite.backprop_rate;
			population[target].bp_fitness = elite.bp_fitness;
			population[target].backprop_momentum = elite.backprop_momentum;
		}
	};

	//prvnich 5 prvku inicializujeme ruznymi variantami blitz a freely growing
	blitz(population[0], 1);
	blitz(population[1], 2);
	freelyGrow(population[2], 1);
	freelyGrow(population[3], 2);
	freelyGrow(population[4], 3);

	//spocitame uvodni fitness funkci
	std::for_each(std::execution::par_unseq, population.begin(), population.end(), [&sample, &samples, &desired](Genome_Type& g) {
		g.fitness = avgQuadError(g.n, sample, 0, samples, desired);
		g.next_fitness = g.fitness;
		cloneNetworkWeights(g.n, g.next_n);
	});

	std::partial_sort(std::execution::par_unseq, population.begin(), population.begin() + (population_size / 2), population.end(), [](const Genome_Type& a, const Genome_Type& b) {
		return a.fitness < b.fitness;
	});

	size_t generation = 0;

	//lambda funkce - krizeni - pres vahy konkretniho neuronu
	auto crossover_neuronweights = [&reng, &layer_gen, &genNumber, &norm_gen](Genome_Type& source, Genome_Type& target) {

		//vygeneruje se nahodna vrstva, neuron a crossover point ve vektoru vah
		size_t x_layer = layer_gen(reng);
		size_t x_neuron = genNumber(0, source.next_n->layers[x_layer].neurons.size());
		size_t x_weight = genNumber(1, source.next_n->layers[x_layer].neurons[x_neuron].weights.size() - 1);
		//vygeneruje se smer krizeni (jestli se bude nahrazovat cast vlevo nebo vpravo od bodu krizeni
		bool direction = (norm_gen(reng) < 0.5) ? true : false;

		//krizeni
		for (size_t w = 0; w < source.next_n->layers[x_layer].neurons[x_neuron].weights.size(); w++) {
			//nalevo nebo napravo od bodu krizeni dle smeru krizeni
			if ((direction && (w < x_weight)) || (!direction && w >= x_weight)) {
				std::swap(target.next_n->layers[x_layer].neurons[x_neuron].weights[w], source.next_n->layers[x_layer].neurons[x_neuron].weights[w]);
			}
		}

	};

	//lambda funkce - krizeni - pres vsechny vahy neuronu a vahy na nej navazujici
	auto crossover_pathway = [&reng, &layer_gen, &genNumber, &norm_gen](Genome_Type& source, Genome_Type& target) {

		size_t x_layer = 0; //TODO: zobecnit pro vice vrstev
		//vygeneruje se neuron z vrstvy
		size_t x_neuron = genNumber(0, source.next_n->layers[x_layer].neurons.size());

		//krizeni vsech vah daneho neuronu
		for (size_t w = 0; w < source.next_n->layers[x_layer].neurons[x_neuron].weights.size(); w++) {
			std::swap(target.next_n->layers[x_layer].neurons[x_neuron].weights[w], source.next_n->layers[x_layer].neurons[x_neuron].weights[w]);
		}

		//pokud nejsme v posledni vrstve udelame totez i se vsemi vahami co nasleduji (vahuji synapse z neuronu do dalsi vrstvy)
		if (x_layer != source.next_n->layers.size() - 1) {

			x_layer++;

			for (size_t n = 0; n < source.next_n->layers[x_layer].neurons.size(); n++) {
				std::swap(target.next_n->layers[x_layer].neurons[n].weights[x_neuron], source.next_n->layers[x_layer].neurons[n].weights[x_neuron]);
			}

		}

	};

	//lambda funkce - najde nejblizsi neuron (podle vstupujicich vah) k zadanemu
	auto findClosestNeuron = [](const NEURON& source, const std::vector<NEURON>& neurons) -> size_t {

		size_t closestIdx = 0;
		double closestDistance = std::numeric_limits<double>::max();

		for (size_t nIdx = 0; nIdx < neurons.size(); nIdx++) {
			double acc = 0.0;
			for (size_t i = 0; i < source.weights.size(); i++) {
				const double tmp = (source.weights[i] - neurons[nIdx].weights[i]);
				acc += tmp*tmp;
			}

			if (acc < closestDistance) {
				closestIdx = nIdx;
				closestDistance = acc;
			}
		}

		return closestIdx;

	};

	//vyber kandidata pro krizeni, vybere se nahodny clen populace bez zadaneho
	auto selectCandidate = [&pop_gen, &reng](size_t except) {
		size_t target;
		do {
			target = pop_gen(reng);
		} while (target == except);
		return target;
	};

	//vyber jednoho z P nejlepsich kandidatu
	auto selectPBestCandidate = [&pop_gen, &reng](size_t except, size_t p) {
		size_t target;
		do {
			target = pop_gen(reng) % p;
		} while (target == except);
		return target;
	};

	//vyber kandidata pro diferencialni krizeni
	auto selectDiffCandidates = [&pop_gen, &reng](size_t except) {
		size_t target1, target2;
		do {
			target1 = pop_gen(reng);
			target2 = pop_gen(reng);
		} while (target1 == except || target2 == except || target1 == target2);

		return std::make_pair(target1, target2);
	};

	//aritmeticky prumer pres vsechny CR
	auto calcCRMean = [&population]() {
		double cr = 0;
		size_t survivors = 0;
		for (auto& p : population) {
			if (p.next_fitness < p.fitness) {
				cr += p.diff_CR;
				survivors++;
			}
		}

		if (survivors == 0) {
			return 0.5; //reset do pocatecniho stavu pokud se nic nezlepsilo
		}

		return cr / static_cast<double>(survivors);
	};

	//Lehmeruv prumer pres vsechny F
	auto calcFMean = [&population]() {
		double f = 0.0;
		double f2 = 0.0;

		for (auto& p : population) {
			if (p.next_fitness < p.fitness) {
				f += p.diff_F;
				f2 += p.diff_F * p.diff_F;
			}
		}

		return f2 / f;
	};

	//vazeny aritmeticky prumer pres vsechny backprop rate tech kteri maji uspesnost uceni pod 1 (zlepsuji se)
	auto calcBPRateMean = [&population]() {
		double bprate = 0;
		double survWeight = 0.0;
		for (auto& p : population) {
			if constexpr (WithBackprop) {
				if (p.bp_fitness < 1.0) {
					bprate += p.backprop_rate * (1.0 - p.bp_fitness);
					survWeight += (1.0 - p.bp_fitness);
				}
			}
		}

		if (survWeight == 0) {
			return 0.5; //reset do pocatecniho stavu pokud se nic nezlepsilo
		}

		return bprate / survWeight;
	};

	//aritmeticky prumer pres vsechny momentum rate tech kteri maji uspesnost uceni pod 1 (zlepsuji se)
	auto calcMomentumMean = [&population]() {
		double momentum = 0;
		double survWeight = 0.0;
		for (auto& p : population) {
			if constexpr (WithBackprop) {
				if (p.bp_fitness < 1.0) {
					momentum += p.backprop_momentum * (1.0 - p.bp_fitness);
					survWeight += (1.0 - p.bp_fitness);
				}
			}
		}

		if (survWeight == 0) {
			return 0.5; //reset do pocatecniho stavu pokud se nic nezlepsilo
		}

		return momentum / survWeight;
	};

	size_t microsAccumulator = 0;

	auto last_checkpoint_time = std::chrono::system_clock::now();

	std::vector<size_t> population_indexes(population_size);
	std::iota(population_indexes.begin(), population_indexes.end(), 0);

	do {

		auto gen_start_time = std::chrono::steady_clock::now();

		//elitism - ulozit nejlepsiho clena populace, kdyby doslo behem generace ke zhorseni, abychom meli v populaci stale historicky nejlepsiho nalezeneho
		saveElite();

		{
			if constexpr (WithBackprop) {

				//50/50 se bud pregeneruji parametry nebo se budou diferencialne krizit
				if (norm_gen(reng) < 0.5) {
					std::for_each(std::execution::par_unseq, population.begin(), population.end(), [&](Genome_Type& g) {
						g.backprop_rate = genBackpropRate();
						g.backprop_momentum = genBackpropRate();
					});
				}
				else {
					std::for_each(std::execution::par_unseq, population.begin(), population.end(), [&](Genome_Type& g) {
						if (norm_gen(reng) < p_meta_crossover) { //krizeni metaparametru, current-to-best

							size_t bestBpFitnessIdx = 0;

							for (size_t p = 0; p < population_size / 4; p++) {
								if (population[p].bp_fitness != 0.0 && population[p].bp_fitness < population[bestBpFitnessIdx].bp_fitness) {
									bestBpFitnessIdx = p;
								}
							}

							g.backprop_rate = g.backprop_rate + g.diff_F * (population[bestBpFitnessIdx].backprop_rate - g.backprop_rate);
							g.backprop_momentum = g.backprop_momentum + g.diff_F * (population[bestBpFitnessIdx].backprop_momentum - g.backprop_momentum);
							g.backprop_epochs = population[bestBpFitnessIdx].backprop_epochs;
						}
					});
				}

				//backpropagation na vsechny s jedinym krokem
				std::for_each(std::execution::par_unseq, population.begin(), population.end(), [&](Genome_Type& g) {
					double bpErr = backpropagation(g.next_n, sample, g.batch_start, g.batch_size, desired, g.backprop_rate, g.backprop_momentum, 1, true, true);
					g.bp_fitness = (g.bp_fitness + bpErr / g.fitness) * 0.5;

					//pokud se chromozom uz nezlepsuje backpropagaci a ma dojit ke zmene metaparametru evoluce, nageneruji se nove
					if (g.bp_fitness > 0.9) {
						if (norm_gen(reng) < p_rate_change) { //zmena metaparametru nahodne
							g.backprop_epochs = genBackpropEpochs();
							g.bp_fitness = 0.0;  //dame mu sanci
						}
					}

					//pokud ma byt clen vyvijen backpropagation algoritmem vice nez ostatni - vetsi sance pro ty, kteri se backpropagation algoritmem zlepsuji lepe
					if (norm_gen(reng) * g.bp_fitness < p_backprop) {

						const size_t epochs = g.backprop_epochs;

						bpErr = backpropagation(g.next_n, sample, g.batch_start, g.batch_size, desired, g.backprop_rate, g.backprop_momentum, epochs, true, true);
						g.bp_fitness = (g.bp_fitness + bpErr / g.fitness) * 0.5;
					}
				});

				const double bpmean = (1.0 - diff_jade_c) * norm_BPRate_gen.mean() + diff_jade_c * calcBPRateMean();
				const double momentummean = (1.0 - diff_jade_c) * norm_Momentum_gen.mean() + diff_jade_c * calcMomentumMean();

				norm_BPRate_gen = std::normal_distribution<double>(bpmean, 0.1);
				norm_Momentum_gen = std::normal_distribution<double>(momentummean, 0.1);
			}

			if constexpr (WithEvolution) {

				std::for_each(std::execution::par_unseq, population_indexes.begin(), population_indexes.end(), [&](size_t i) {

					if (norm_gen(reng) < p_rate_change) {
						population[i].diff_strategy = genDEStrategy();
						population[i].batch_size = genBatchSize();
						population[i].batch_start = genBatchStart(population[i].batch_size);
					}

					population[i].diff_F = genFParameter();
					population[i].diff_CR = genCRParameter();

					//krizeni
					{
						const double crossover_chance = norm_gen(reng);

						selectRand(crossover_chance,
							//krizeni pres vahy jednoho neuronu
							p_crossover_weights,
								[&crossover_neuronweights, &population, &selectCandidate, i]() {
									size_t target = selectCandidate(i);

									size_t source = population[i].fitness < population[target].fitness ? i : target;
									target = population[i].fitness < population[target].fitness ? target : i;

									crossover_neuronweights(population[source], population[target]);
								},
							//krizeni pres cesty vah jednoho neuronu
							p_crossover_pathway,
								[&crossover_pathway, &population, &selectCandidate, i]() {
									size_t target = selectCandidate(i);

									size_t source = population[i].fitness < population[target].fitness ? i : target;
									target = population[i].fitness < population[target].fitness ? target : i;

									crossover_pathway(population[source], population[target]);
								},
							//diferencialni krizeni
							p_crossover_diff,
								[&crossover_pathway, &population, &selectCandidate, &selectDiffCandidates, &selectPBestCandidate, &findClosestNeuron, &elite, &norm_gen, &reng, i]() {

									bool findBestNeuron = (norm_gen(reng) < 0.5);

									switch (population[i].diff_strategy) {

										case DECrossoverStrategy::Rand_1:
										{
											auto source = selectCandidate(i);
											auto sources = selectDiffCandidates(i);

											for (size_t x_layer = 0; x_layer < population[i].next_n->layers.size(); x_layer++) {
												for (size_t x_neuron = 0; x_neuron < population[i].next_n->layers[x_layer].neurons.size(); x_neuron++) {

													size_t y1_neuron = findBestNeuron ? findClosestNeuron(population[source].n->layers[x_layer].neurons[x_neuron], population[sources.first].next_n->layers[x_layer].neurons) : x_neuron;
													size_t y2_neuron = findBestNeuron ? findClosestNeuron(population[sources.first].n->layers[x_layer].neurons[y1_neuron], population[sources.second].next_n->layers[x_layer].neurons) : x_neuron;

													for (size_t x_weight = 0; x_weight < population[i].next_n->layers[x_layer].neurons[x_neuron].weights.size(); x_weight++) {

														if (norm_gen(reng) < population[i].diff_CR) {

															population[i].next_n->layers[x_layer].neurons[x_neuron].weights[x_weight] =
																population[source].n->layers[x_layer].neurons[x_neuron].weights[x_weight] +
																population[i].diff_F * (population[sources.first].n->layers[x_layer].neurons[y1_neuron].weights[x_weight] - population[sources.second].n->layers[x_layer].neurons[y2_neuron].weights[x_weight]);
														}

													}
												}
											}
											break;
										}
										case DECrossoverStrategy::Best_1:
										{
											auto sources = selectDiffCandidates(i);

											for (size_t x_layer = 0; x_layer < population[i].next_n->layers.size(); x_layer++) {
												for (size_t x_neuron = 0; x_neuron < population[i].next_n->layers[x_layer].neurons.size(); x_neuron++) {

													size_t y1_neuron = findBestNeuron ? findClosestNeuron(elite.n->layers[x_layer].neurons[x_neuron], population[sources.first].next_n->layers[x_layer].neurons) : x_neuron;
													size_t y2_neuron = findBestNeuron ? findClosestNeuron(elite.n->layers[x_layer].neurons[x_neuron], population[sources.second].next_n->layers[x_layer].neurons) : x_neuron;

													for (size_t x_weight = 0; x_weight < population[i].next_n->layers[x_layer].neurons[x_neuron].weights.size(); x_weight++) {

														if (norm_gen(reng) < population[i].diff_CR) {

															population[i].next_n->layers[x_layer].neurons[x_neuron].weights[x_weight] =
																elite.n->layers[x_layer].neurons[x_neuron].weights[x_weight] +
																population[i].diff_F * (population[sources.first].n->layers[x_layer].neurons[y1_neuron].weights[x_weight] - population[sources.second].n->layers[x_layer].neurons[y2_neuron].weights[x_weight]);
														}

													}
												}
											}
											break;
										}
										case DECrossoverStrategy::Cur_To_Best_1:
										{
											auto sources = selectDiffCandidates(i);

											for (size_t x_layer = 0; x_layer < population[i].next_n->layers.size(); x_layer++) {
												for (size_t x_neuron = 0; x_neuron < population[i].next_n->layers[x_layer].neurons.size(); x_neuron++) {

													size_t y_neuron = findBestNeuron ? findClosestNeuron(elite.n->layers[x_layer].neurons[x_neuron], population[i].next_n->layers[x_layer].neurons) : x_neuron;
													size_t y1_neuron = findBestNeuron ? findClosestNeuron(elite.n->layers[x_layer].neurons[x_neuron], population[sources.first].next_n->layers[x_layer].neurons) : x_neuron;
													size_t y2_neuron = findBestNeuron ? findClosestNeuron(elite.n->layers[x_layer].neurons[x_neuron], population[sources.second].next_n->layers[x_layer].neurons) : x_neuron;

													for (size_t x_weight = 0; x_weight < population[i].next_n->layers[x_layer].neurons[x_neuron].weights.size(); x_weight++) {

														if (norm_gen(reng) < population[i].diff_CR) {

															population[i].next_n->layers[x_layer].neurons[x_neuron].weights[x_weight] =
																population[i].n->layers[x_layer].neurons[x_neuron].weights[x_weight] +
																population[i].diff_F * (elite.n->layers[x_layer].neurons[x_neuron].weights[x_weight] - population[i].n->layers[x_layer].neurons[y_neuron].weights[x_weight]) +
																population[i].diff_F * (population[sources.first].n->layers[x_layer].neurons[y1_neuron].weights[x_weight] - population[sources.second].n->layers[x_layer].neurons[y2_neuron].weights[x_weight]);
														}

													}
												}
											}
											break;
										}
										case DECrossoverStrategy::Cur_To_Rand_1:
										{
											auto source_rand = selectCandidate(i);
											auto sources = selectDiffCandidates(i);

											for (size_t x_layer = 0; x_layer < population[i].next_n->layers.size(); x_layer++) {
												for (size_t x_neuron = 0; x_neuron < population[i].next_n->layers[x_layer].neurons.size(); x_neuron++) {

													size_t y_neuron = findBestNeuron ? findClosestNeuron(population[i].n->layers[x_layer].neurons[x_neuron], population[source_rand].next_n->layers[x_layer].neurons) : x_neuron;
													size_t y1_neuron = findBestNeuron ? findClosestNeuron(population[source_rand].n->layers[x_layer].neurons[y_neuron], population[sources.first].next_n->layers[x_layer].neurons) : x_neuron;
													size_t y2_neuron = findBestNeuron ? findClosestNeuron(population[sources.first].n->layers[x_layer].neurons[y1_neuron], population[sources.second].next_n->layers[x_layer].neurons) : x_neuron;

													for (size_t x_weight = 0; x_weight < population[i].next_n->layers[x_layer].neurons[x_neuron].weights.size(); x_weight++) {

														if (norm_gen(reng) < population[i].diff_CR) {

															population[i].next_n->layers[x_layer].neurons[x_neuron].weights[x_weight] =
																population[i].n->layers[x_layer].neurons[x_neuron].weights[x_weight] +
																population[i].diff_F * (population[source_rand].n->layers[x_layer].neurons[y_neuron].weights[x_weight] - population[i].n->layers[x_layer].neurons[y1_neuron].weights[x_weight]) +
																population[i].diff_F * (population[sources.first].n->layers[x_layer].neurons[y1_neuron].weights[x_weight] - population[sources.second].n->layers[x_layer].neurons[y2_neuron].weights[x_weight]);
														}

													}
												}
											}
											break;
										}
										case DECrossoverStrategy::Cur_To_PBest_1:
										{
											auto sources = selectDiffCandidates(i);
											auto pbest = selectPBestCandidate(i, population[i].diff_P);

											for (size_t x_layer = 0; x_layer < population[i].next_n->layers.size(); x_layer++) {
												for (size_t x_neuron = 0; x_neuron < population[i].next_n->layers[x_layer].neurons.size(); x_neuron++) {

													size_t y_neuron = findBestNeuron ? findClosestNeuron(population[pbest].n->layers[x_layer].neurons[x_neuron], population[i].next_n->layers[x_layer].neurons) : x_neuron;
													size_t y1_neuron = findBestNeuron ? findClosestNeuron(population[pbest].n->layers[x_layer].neurons[x_neuron], population[sources.first].next_n->layers[x_layer].neurons) : x_neuron;
													size_t y2_neuron = findBestNeuron ? findClosestNeuron(population[pbest].n->layers[x_layer].neurons[x_neuron], population[sources.second].next_n->layers[x_layer].neurons) : x_neuron;

													for (size_t x_weight = 0; x_weight < population[i].next_n->layers[x_layer].neurons[x_neuron].weights.size(); x_weight++) {

														if (norm_gen(reng) < population[i].diff_CR) {

															population[i].next_n->layers[x_layer].neurons[x_neuron].weights[x_weight] =
																population[i].n->layers[x_layer].neurons[x_neuron].weights[x_weight] +
																population[i].diff_F * (population[pbest].n->layers[x_layer].neurons[x_neuron].weights[x_weight] - population[i].n->layers[x_layer].neurons[y_neuron].weights[x_weight]) +
																population[i].diff_F * (population[sources.first].n->layers[x_layer].neurons[y1_neuron].weights[x_weight] - population[sources.second].n->layers[x_layer].neurons[y2_neuron].weights[x_weight]);
														}

													}
												}
											}
											break;
										}
									}

								}
						);
					}

					//pokud ma dojit k nahodne mutaci
					if (norm_gen(reng) < p_mutation) {

						//vybere se nahodne vrstva, neuron a jeho jedna vaha a vygeneruje se znovu
						const size_t x_layer = layer_gen(reng);
						const size_t x_neuron = genNumber(0, population[i].next_n->layers[x_layer].neurons.size());
						const size_t x_weight = genNumber(1, population[i].next_n->layers[x_layer].neurons[x_neuron].weights.size() - 1);

						population[i].next_n->layers[x_layer].neurons[x_neuron].weights[x_weight] = drand();
					}

					//dolni cast populace (horsi) se mohou obcas pregenerovat pomoci blitz nebo freelyGrowing
					if (i >= population_size / 2) {

						const double num = norm_gen(reng);

						selectRand(num,
							//nahrazeni metodou Blitz
							p_blitz,
								[&blitz, &population, i]() {
									blitz(population[i], RandomMode);
								},
							//nahrazeni metodou Freely Growing
							p_freely,
								[&freelyGrow, &population, i]() {
									freelyGrow(population[i], RandomMode);
								},
							//nahrazeni nahodnymi vahami
							p_randomgen,
								[&randomReinit, &population, i]() {
									randomReinit(population[i]);
								},
							//nahrazeni elitou
							p_elitereinit,
								[&restoreElite, i]() {
									restoreElite(i);
								}
						);
					}

				});
			}

		}

		//prepocitani fitness funkce a serazeni horni casti populace od nejlepsiho po nejhorsi
		calcPopulationFitness();

		//aktualizace parametru CR a F
		const double nmean_cr = (1.0 - diff_jade_c) * norm_CR_gen.mean() + diff_jade_c * calcCRMean();
		const double nmean_f = (1.0 - diff_jade_c) * cauchy_F_gen.a() + diff_jade_c * calcFMean();

		norm_CR_gen = std::normal_distribution<double>(nmean_cr, 0.1);
		cauchy_F_gen = std::cauchy_distribution<double>(nmean_f, 0.1);

		selection();
		sortPopulation();

		//pokud doslo ke zhorseni nejlepsiho kandidata, nahradi se nejhorsi kandidat ulozenou elitou
		if (population[0].fitness > elite.fitness) {
			restoreElite(0);
		}

		auto gen_end_time = std::chrono::steady_clock::now();

		auto diff = std::chrono::duration_cast<std::chrono::milliseconds>(gen_end_time - gen_start_time);
		microsAccumulator += std::chrono::duration_cast<std::chrono::microseconds>(gen_end_time - gen_start_time).count();

		double mdiff = 0.0;
		for (size_t i = 0; i < population_size; i++) {
			mdiff += population[i].fitness;
		}
		mdiff /= static_cast<double>(population_size);

		double stddev = 0.0;
		for (size_t i = 0; i < population_size; i++) {
			stddev += std::pow(population[i].fitness - mdiff, 2);
		}
		stddev /= static_cast<double>(population_size);
		stddev = std::sqrt(stddev);

		std::cout << "Generace " << generation << ", elite fitness: " << population[0].fitness << ", prum: " << mdiff << "+-" << stddev;
		if constexpr (WithBackprop) {
			std::cout << ", rate: " << population[0].backprop_rate << ", momentum: " << population[0].backprop_momentum << ", epochs: " << population[0].backprop_epochs
				<< ", bpFitness: " << population[0].bp_fitness;
		}
		std::cout << ", gen. time: " << diff.count() << "ms" << std::endl;

		gPopulationStats.push_back({ generation, microsAccumulator, population[0].fitness, mdiff });

		//ulozi nejlepsi vahy do souboru kazdych 50 generaci nebo kazdou minutu pro pripad spadnuti nebo ukonceni
		if (((generation > 0 && (generation % 50) == 0) || std::chrono::duration_cast<std::chrono::minutes>(std::chrono::system_clock::now() - last_checkpoint_time).count() > 1) && population[0].fitness < lastSavedFitness) {
			saveTempWeights(population[0].n);
			lastSavedFitness = population[0].fitness;
			last_checkpoint_time = std::chrono::system_clock::now();
		}

	} while (generation++ < generations_count_min);

	//do n se zkopiruje nove nejlepsi reseni
	cloneNetworkWeights(population[0].n, n);

	return population[0].fitness;
}
