#pragma once

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <cmath>
#include <fstream>
#include <string>
#include <vector>
#include <memory>
#include <iostream>
#include <execution>
#include <random>
#include <functional>

#undef min

//podminka ukonceni zpetneho sireni kdyz je prumerna kvadraticka chyba mensi nez EPSILON
constexpr double EPSILON = 0.00000005;

//vycet druhu uceni
enum class LearnType {
	None,							//bez dalsiho uceni
	Backpropagation,				//klasicky backprop
	Genetic_Algorithm_Backprop,		//diferencialni evoluce s backprop
	Genetic_Algorithm_NoBackprop,	//diferencialni evoluce bez backprop
	Meta_Backprop,					//metaheuristicka diferencialni evoluce backpropagation
};

//struktura neuronu
struct NEURON {
	std::vector<double> weights;		//vstupni vahy neuronu
	std::vector<double> last_weights;	//vahy neuronu z minule iterace
};

//struktura vrstvy
struct LAYER {
	std::vector<NEURON> neurons;		//neurony vrstvy
};

//struktura neuronove site vicevrstveho perceptronu
struct NETWORK {
	size_t inputs = 0;					//pocet vstupnich priznaku
	std::vector<LAYER> layers;			//pole vrstev
};

//struktura na uchovani informaci o postupu konvergence
struct PopulationStatistics {
	size_t generation;			//aktualni generace
	size_t microseconds;		//cas od pocatku behu
	double bestFitness;			//chyba nejlepsiho prvku
	double avgFitness;			//prumerna chyba
};

//pole statistik konvergence
extern std::vector<PopulationStatistics> gPopulationStats;

//pole predaktivaci pro kazde vlakno
extern thread_local std::vector<std::vector<double>> tlPreactivations;
//pole aktivaci pro kazde vlakno
extern thread_local std::vector<std::vector<double>> tlActivations;

//vytvori sit se zadanymi parametry
//inputs = pocet vstupnich priznaku, layers = pocet vrstev, layerNeurons = pocet neuronu v kazde vrstve
std::unique_ptr<NETWORK> createNetwork(size_t inputs, size_t layers, std::vector<int>& layerNeurons, bool fromLoadedWeights = false);
//duplikuje sit a podle hodnoty parametru randWeights ji prepise nahodnymi vahami
std::unique_ptr<NETWORK> cloneNetwork(std::unique_ptr<NETWORK>& n, bool randWeights);
//duplikuje vahy jedne site do druhe
void cloneNetworkWeights(std::unique_ptr<NETWORK>& n, std::unique_ptr<NETWORK>& target);

//vypocita aktivace dane site pro zadany vzorek, pouziva k tomu pole tlPreactivations a tlActivations v jednom vlakne
void calculateNetworkActivations(std::unique_ptr<NETWORK>& n, const std::vector<double>& sample);

//vypocita kvadratickou odchylku pri klasifikaci vzorku sample oproti odpovedi ucitele desired
double quadraticError(std::unique_ptr<NETWORK>& n, const std::vector<double>& sample, const std::vector<double>& desired);
//vypocita prumernou kvadratickou odchylku pro zadane prvky
double avgQuadError(std::unique_ptr<NETWORK>& n, const std::vector<std::vector<double>>& samples, size_t startIndex, size_t count, const std::vector<std::vector<double>>& desired);

//ziska gradienty pro dany vzorek
void getDelta(std::unique_ptr<NETWORK>& n, const std::vector<double>& sample, const std::vector<double>& desired, size_t layer, std::vector<std::vector<double>>& deltabuffers);

//upravi vahy neuronu podle vypoctenych gradientu, adjustSourceActivation je bud aktivace predchozi vrstvy nebo vzorek
void adjustWeights(std::unique_ptr<NETWORK>& n, size_t layer, const std::vector<double>& adjustSourceActivation, const std::vector<double>& delta, double learningRate, double momentumRate);

//vypocita aktivacni funkci (logisticka sigmoida) pro jedno cislo
double sigmoid(double x);
//vypocita derivaci aktivacni funkce pro jedno cislo
double activationDeriv(double x);

//vygeneruje nahodne cislo pro vahu synapse
double drand();

//zkontroluje, naparsuje a pouzije argumenty z prikazove radky
void checkArgs(int argc, char** argv, int& inputs, int& outputs, int& samples, int& layers, std::vector<int>& npl, int& initW, std::vector<double>& wValues, int& showEqms, int& showEpochs, int& notShowWeights);

//ulozi vahy site do zadaneho souboru
void saveWeights(std::unique_ptr<NETWORK>& n, const std::string& outputModelFile);
//ulozi statistiky populace do souboru
void savePopulationStats();
//ulozi mezivysledek vah do souboru
void saveTempWeights(std::unique_ptr<NETWORK>& n);
//vytiskne na obrazovku navod k pouziti
void showUsage();

//inicializuje sit metodou Blitz
void blitzInit(std::unique_ptr<NETWORK>& network, const std::vector<std::vector<double>>& sample, const std::vector<std::vector<double>>& desired, size_t mode);
//inicializuje sit metodou FreelyGrowing
void freelyGrowingInit(std::unique_ptr<NETWORK>& network, const std::vector<std::vector<double>>& sample, const std::vector<std::vector<double>>& desired, size_t mode);

//spusti algoritmus zpetneho sireni nad danou neuronovou siti
double backpropagation(std::unique_ptr<NETWORK>& n, const std::vector<std::vector<double>>& sample, size_t startIndex, size_t count, const std::vector<std::vector<double>>& desired, double learningRate, double momentumRate, size_t epochLimit = 0, bool silent = false, bool allowDiverge = false);
